use std::{collections::HashMap, sync::Mutex};

use crate::models::behavior::BehaviorWithSignature;

lazy_static! {
    static ref BEHAVIOR_POOL: Mutex<HashMap<u64, BehaviorWithSignature>> = Mutex::new({
        let mut m = HashMap::new();
        m.insert(0, BehaviorWithSignature::default());
        m
    });
}

pub fn add_behavior_to_pool(
    behavior: BehaviorWithSignature,
    key: u64,
) -> Option<BehaviorWithSignature> {
    let mut final_key = key;
    let mut behavior_pool = BEHAVIOR_POOL.lock().unwrap();
    while behavior_pool.contains_key(&final_key) {
        final_key += 1;
    }
    let res = behavior_pool.insert(final_key, behavior);
    drop(behavior_pool);
    res
}

pub fn query_behavior_by_id(key: u64) -> Option<BehaviorWithSignature> {
    let behavior_pool = BEHAVIOR_POOL.lock().unwrap();
    behavior_pool.get(&key).cloned()
}

pub fn get_and_remove_behavior_from_pool(key: u64) -> Option<BehaviorWithSignature> {
    let mut behavior_pool = BEHAVIOR_POOL.lock().unwrap();
    let result = behavior_pool.remove(&key);
    drop(behavior_pool);
    result
}

pub fn get_many_behaviors_and_remove_them(
    behaviors_nb_wanted: usize,
) -> Option<Vec<BehaviorWithSignature>> {
    let mut behavior_pool = BEHAVIOR_POOL.lock().unwrap();
    if behavior_pool.len() < behaviors_nb_wanted {
        return Option::None;
    }
    let mut behaviors: Vec<BehaviorWithSignature> = Vec::with_capacity(behaviors_nb_wanted);
    let keys: Vec<_> = {
        behavior_pool
            .keys()
            .take(behaviors_nb_wanted)
            .copied()
            .collect()
    };
    for k in keys {
        behaviors.push(behavior_pool.remove(&k).unwrap());
    }
    Some(behaviors)
}

#[cfg(test)]
mod test {
    use crate::{
        accessors::behavior_pool::{
            add_behavior_to_pool, get_and_remove_behavior_from_pool,
            get_many_behaviors_and_remove_them, BEHAVIOR_POOL,
        },
        models::behavior::BehaviorWithSignature,
    };

    #[test]
    fn insert_and_retrieve_behavior() {
        let new_behavior = BehaviorWithSignature::default();
        add_behavior_to_pool(new_behavior.clone(), 1);
        assert_eq!(get_and_remove_behavior_from_pool(1).unwrap(), new_behavior);
    }

    #[test]
    fn retrieve_many_behaviors() {
        let new_behavior = BehaviorWithSignature::default();
        let new_behavior_2 = BehaviorWithSignature::default();
        let new_behavior_3 = BehaviorWithSignature::default();
        let new_behavior_4 = BehaviorWithSignature::default();
        let new_behavior_5 = BehaviorWithSignature::default();
        let new_behavior_6 = BehaviorWithSignature::default();
        let new_behavior_7 = BehaviorWithSignature::default();
        add_behavior_to_pool(new_behavior, 1);
        add_behavior_to_pool(new_behavior_2, 2);
        add_behavior_to_pool(new_behavior_3, 3);
        add_behavior_to_pool(new_behavior_4, 4);
        add_behavior_to_pool(new_behavior_5, 5);
        add_behavior_to_pool(new_behavior_6, 6);
        add_behavior_to_pool(new_behavior_7, 7);

        let many_behaviors = get_many_behaviors_and_remove_them(7).unwrap();
        assert_eq!(many_behaviors.len(), 7);
        let mut pool = BEHAVIOR_POOL.lock().unwrap();
        assert_eq!(pool.len(), 1);
        pool.clear();
        drop(pool);
    }
}
