use std::{collections::HashMap, sync::Mutex};

use hex_literal::hex;

use crate::models::*;

lazy_static! {
    static ref TRANSACTION_POOL: Mutex<HashMap<u64, MessageWithSignature>> = Mutex::new({
        let mut m = HashMap::new();
        let random_message = MessageWithSignature {
            message: Message::EIP2930 {
                chain_id: ChainId(5),
                nonce: 7,
                gas_price: 30_000_000_000_u64.into(),
                gas_limit: 5_748_100_u64,
                action: TransactionAction::Call(
                    hex!("811a752c8cd697e3cb27279c330ed1ada745a8d7").into(),
                ),
                value: 2.as_u256() * 1_000_000_000 * 1_000_000_000,
                input: hex!("6ebaf477f83e051589c1188bcc6ddccd").to_vec().into(),
                access_list: vec![
                    AccessListItem {
                        address: hex!("de0b295669a9fd93d5f28d9ec85e40f4cb697bae").into(),
                        slots: vec![
                            hex!(
                                "0000000000000000000000000000000000000000000000000000000000000003"
                            )
                            .into(),
                            hex!(
                                "0000000000000000000000000000000000000000000000000000000000000007"
                            )
                            .into(),
                        ],
                    },
                    AccessListItem {
                        address: hex!("bb9bc244d798123fde783fcc1c72d3bb8c189413").into(),
                        slots: vec![],
                    },
                ],
            },
            signature: MessageSignature::new(
                false,
                hex!("36b241b061a36a32ab7fe86c7aa9eb592dd59018cd0443adc0903590c16b02b0"),
                hex!("5edcc541b4741c5cc6dd347c5ed9577ef293a62787b4510465fadbfe39ee4094"),
            )
            .unwrap(),
        };
        m.insert(0, random_message);
        m
    });
}

pub fn add_transaction_to_pool(
    transaction: MessageWithSignature,
    key: u64,
) -> Option<MessageWithSignature> {
    let mut final_key = key;
    let mut transaction_pool = TRANSACTION_POOL.lock().unwrap();
    while transaction_pool.contains_key(&final_key) {
        final_key += 1;
    }
    transaction_pool.insert(final_key, transaction)
}

pub fn query_transaction_by_id(key: u64) -> Option<MessageWithSignature> {
    let transaction_pool = TRANSACTION_POOL.lock().unwrap();
    transaction_pool.get(&key).cloned()
}

pub fn get_and_remove_transaction_from_pool(key: u64) -> Option<MessageWithSignature> {
    let mut transaction_pool = TRANSACTION_POOL.lock().unwrap();
    let result = transaction_pool.remove(&key);
    drop(transaction_pool);
    result
}

pub fn get_many_transactions_and_remove_them(
    transactions_nb_wanted: usize,
) -> Option<Vec<MessageWithSignature>> {
    let mut transaction_pool = TRANSACTION_POOL.lock().unwrap();
    if transaction_pool.len() < transactions_nb_wanted {
        return Option::None;
    }
    let mut transactions: Vec<MessageWithSignature> = Vec::with_capacity(transactions_nb_wanted);
    let keys: Vec<_> = {
        transaction_pool
            .keys()
            .take(transactions_nb_wanted)
            .copied()
            .collect()
    };
    for k in keys {
        transactions.push(transaction_pool.remove(&k).unwrap());
    }
    Some(transactions)
}
